
#ifndef lint
static char sccsid[] = "@(#)qlalloc.c 3.2 4/10/90 Arthur van Hoff (Turing Institute)";
#endif lint


/************************************************************************
 * Quick List allocation (version 1.2), (c)Turing Institute 1989
 * By: Arthur van Hoff, based on the algorithm by Charles B. Weinstock and 
 * William A. Wulf.
 *  
 * qlalloc() provides fast allocation of blocks of data. For each block size 
 * upto QLMAX a list of free blocks of that particular size if maintained. 
 * When a such a block is requested this appropriate list is checked. 
 * If the list is not empty the head of * the list is returned. If the list 
 * is empty a block is allocated from the heap.
 * If the size of the allocate block is greater or equal to QLMAX the qlmisc list 
 * is checked. This list contains all free blocks with size greater or equal to QLMAX.
 * If the block size is not found in the qlmisc list the block is taken from the heap.
 * The heap is allocate using malloc.
 *
 * qlfree() will put blocks smaller than QLMAX into the appropriate free list. 
 * All other blocks will be put into the qlmisc list.
 *
 *
 */
 
#include <stdio.h>
#include "qlalloc.h"

/************************************************************************
 * The following flags can be defined:
 *	QLZERO		- qlalloc erases the contents of the block for you
 *	QLCHECK 	- qlalloc/qlfree will perform additional checks wether
 *			  to see wether you use them correctly. Some additional
 *			  admin info is added to the end of each block.
 *	QLSTATISTICS	- qlalloc/qlfree will keep some statistics which can
 *			  be printed using qlstatistics().
 *	QLUSAGE		- qlalloc will keep record of the number of times
 *			  blocks are allocated. This is usefull if you want
 *			  to find out which block sizes are used most frequently.
 */
 
/*
#define	QLUSAGE		1
#define	QLCHECK		1
#define	QLSTATISTICS	1
*/
#define	QLZERO		1

/************************************************************************
 * The following constants are defined:
 *	QLMAX		- Maximum byte size of blocks in the quick-lists.
 *	QLMIN		- Minimum byte size of blocks in the quick-lists (at least 4!).
 *	QLBLOCKSIZE	- Default size of a heap block.
 *	QLMINBLOCKSIZE	- Minimal size of a heap block.
 */

#define QLMAX		((QLTYPE)512)
#define QLMIN		((QLTYPE)sizeof(struct ql_list *))

#define	QLBLOCKSIZE	(QLMAX * 64)
#define	QLMINBLOCKSIZE	(QLMAX * 16)

/************************************************************************/

#define	QLMARGIN(bytes)	((bytes) + ((bytes) >> 3))
#define	QLEXTRA		sizeof(union ql_header)
#define	QLROUND(bytes)	(((bytes) <= QLMIN) ?\
				 QLMIN : (((bytes) + (QLALIGN - 1)) & ~(QLALIGN - 1)))

extern char *malloc();
#define	ALLOCMEM(bytes)	((void *)malloc((unsigned)(bytes)))

struct ql_list {
	union ql_header		header;
	struct ql_list		*next;
};

/************************************************************************/

#ifdef QLCHECK

struct ql_trailer {
	QLTYPE		magic;
	struct ql_list	*next;
};

struct ql_list *qlblocks = 0L;

#undef QLEXTRA
#define QLEXTRA		(sizeof(union ql_header) + sizeof(struct ql_trailer))
#define	QLMAGICA	((QLTYPE)-1)
#define	QLMAGICF	((QLTYPE)-2)
#define	qltrailer(p)	((struct ql_trailer *)((long)(p) + ((struct ql_list *)(p))->header.bytes + sizeof(union ql_header)))

#endif QLCHECK

/************************************************************************/

#ifdef QLSTATISTICS
long qlheapmem, qlusedmem, qlused;
#endif QLSTATISTICS

#ifdef QLUSAGE
long *qluseage, qlmiscuse;
#endif QLUSAGE

struct ql_list **qllist = NULL, *qlmisc;
char	*qlheap;
QLTYPE	qlheapsize = 0;

/************************************************************************/

void *_qlalloc(bytes)
register QLTYPE bytes;
{
	register struct ql_list *new;

	if (!qllist)
		qlinit();
#ifdef QLCHECK
	if (bytes <= 0)
		fprintf(stderr,"warning: qlalloc, allocating 0 bytes\n");
#endif QLCHECK
	bytes = QLROUND(bytes);
again:
	if (bytes < QLMAX) {
		register struct ql_list **p = &qllist[bytes/QLALIGN];
#ifdef QLUSAGE
		qluseage[bytes/QLALIGN]++;
#endif QLUSAGE
		if (new = *p)
			*p = new->next;
		else {
			if ((bytes + QLEXTRA) > qlheapsize) {
#ifdef QLCHECK
				fprintf(stderr,"*** CBREAK ***\n");
#endif QLCHECK
				/* allocate more heap space */
				if (qlheapsize >= (QLMIN + QLEXTRA)) {
					/* free remaining space */
					new = (struct ql_list *)qlheap;
					new->header.bytes = qlheapsize - QLEXTRA;
#ifdef QLSTATISTICS
					qlused++;
					qlusedmem += new->header.bytes;
#endif QLSTATISTICS
#ifdef QLCHECK
					qltrailer(new)->magic = QLMAGICA;
					qltrailer(new)->next = qlblocks;
					qlblocks = new;
#endif QLCHECK
					qlfree(&new->next);
				}
				/* Check if the largest block on the misc list
				 * is big enough to serve as heap space.
				 */
				if (qlmisc && (qlmisc->header.bytes >= QLMINBLOCKSIZE)) {
#ifdef QLCHECK
					register struct ql_list **p = &qlblocks;
					
					/* take the block out of the block list */
					while (*p && (*p != qlmisc))
						p = &qltrailer(*p)->next;
					if (*p == qlmisc)
						*p = qltrailer(*p)->next;
					else
						fprintf(stderr,"*** qlalloc: bad block ***\n");
#endif QLCHECK
					/* take heap space of the misc list */
					qlheapsize = qlmisc->header.bytes + QLEXTRA;
					qlheap = (char *)qlmisc;
					qlmisc = qlmisc->next;
				} else {
					/* allocate heap space */
					if (!(qlheap = (char *)ALLOCMEM(QLBLOCKSIZE)))
						return(0L);
					qlheapsize = QLBLOCKSIZE;
#ifdef QLSTATISTICS
					qlheapmem += qlheapsize;
#endif QLSTATISTICS
				}
			}
			new = (struct ql_list *)qlheap;
			new->header.bytes = bytes;
#ifdef QLCHECK
			qltrailer(new)->next = qlblocks;
			qlblocks = new;
#endif QLCHECK
			qlheap += bytes + QLEXTRA;
			qlheapsize -= bytes + QLEXTRA;
		}
	} else {
		register struct ql_list **p = &qlmisc, **fit = 0L;
#ifdef QLUSAGE
		qlmiscuse++;
#endif QLUSAGE
		while (*p && ((*p)->header.bytes > bytes))
			p = &(*(fit = p))->next;
		
		if ((new = *p) && (new->header.bytes == bytes))
			*p = new->next;
		else if (fit && (new = *fit) && (new->header.bytes <= QLMARGIN(bytes)))
			*fit = new->next;
		else {
			if (!(new = (struct ql_list *)ALLOCMEM(bytes + QLEXTRA))) {
#ifdef QLCHECK
				fprintf(stderr,"*** qlalloc: out of memory ***\n");
#endif QLCHECK
				return(0L);
			}
#ifdef QLSTATISTICS
			qlheapmem += bytes + QLEXTRA;
#endif QLSTATISTICS
			new->header.bytes = bytes;
#ifdef QLCHECK
			qltrailer(new)->next = qlblocks;
			qlblocks = new;
#endif QLCHECK
		}
	}
#ifdef QLSTATISTICS
	qlusedmem += new->header.bytes;
	qlused++;
#endif QLSTATISTICS
#ifdef QLCHECK
	qltrailer(new)->magic = QLMAGICA;
#endif QLCHECK
#ifdef QLZERO
	memset(&new->next,0,new->header.bytes);
#endif QLZERO
	return((void *)&new->next);
}

/************************************************************************/

void qlfree(p)
register union ql_header *p;
{
	register QLTYPE bytes;
	register struct ql_list *new;

#ifdef QLCHECK
	if (!p) {
		fprintf(stderr,"*** qlfree: freeing NULL ***\n");
		return;
	}
#endif QLCHECK
	bytes = (--p)->bytes;
	new = (struct ql_list *)p;
#ifdef QLCHECK
	if (bytes <= 0) {
		fprintf(stderr,"*** qlfree: bad block size ***\n");
		return;
	}
#ifdef QLSTATISTICS
	if (qlused <= 0) {
		fprintf(stderr,"*** qlfree: freeing block when no blocks are allocated ***\n");
		return;
	}
#endif QLSTATISTICS
	if (qltrailer(new)->magic == QLMAGICF) {
		fprintf(stderr,"*** qlfree: freeing free block ***\n");
		return;
	}
	if (qltrailer(new)->magic != QLMAGICA) {
		fprintf(stderr,"*** qlfree: bad block ***\n");
		return;
	}
	qltrailer(new)->magic = QLMAGICF;
#endif QLCHECK
#ifdef QLSTATISTICS
	qlusedmem -= bytes;
	qlused--;
#endif QLSTATISTICS
	if (bytes < QLMAX) {
		new->next = qllist[bytes / QLALIGN];
		qllist[bytes / QLALIGN] = new;
	} else {
		register struct ql_list **q = &qlmisc;
		
		while (*q && ((*q)->header.bytes > bytes))
			q = &(*q)->next;
		new->next = *q;
		*q = new;
	}
}

/************************************************************************/

void *qlcopy(p)
void *p;
{
	void *new;
	
#ifdef QLCHECK
	if (!p) {
		fprintf(stderr,"*** qlcopy: NULL block ***\n");
		return(0L);
	}
	if (qlsize(p) <= 0) {
		fprintf(stderr,"*** qlcopy: bad block size ***\n");
		return(0L);
	}
	if (qltrailer(((union ql_header *)p) - 1)->magic != QLMAGICA) {
		fprintf(stderr,"*** qlcopy: bad block ***\n");
		return(0L);
	}
#endif QLCHECK
	if (new = qlalloc(qlsize(p))) {
		memcpy(new,p,qlsize(p));
		return(new);
	} else
		return(0L);
}

/************************************************************************/

int _qlsetsize(p,size)
register void **p;
QLTYPE	size;
{
	QLTYPE bytes;
	void *new;
#ifdef QLCHECK
	if (!*p) {
		fprintf(stderr,"*** qlsetsize: NULL block ***\n");
		return;
	}
	if (qlsize(*p) <= 0) {
		fprintf(stderr,"*** qlsetsize: bad block size ***\n");
		return;
	}
	if (qltrailer(((union ql_header *)*p) - 1)->magic != QLMAGICA) {
		fprintf(stderr,"*** qlsetsize: bad block ***\n");
		return;
	}
#endif QLCHECK

	bytes = qlsize(*p);
	if (new = qlalloc(size)) {
		memcpy(new,*p,(bytes < size) ? bytes : size);
		qlfree(*p);
		*p = new;
		return(1);
	}
	return(0);
}

/************************************************************************/

qlstatistics()
{
#ifdef QLCHECK
	register struct ql_list *p = qlblocks;
	register long	n = 0, b = 0;
	
	for ( ; p ; p = qltrailer(p)->next, n++) {
		if ((qltrailer(p)->magic != QLMAGICA) && (qltrailer(p)->magic != QLMAGICF)) {
			fprintf(stderr,"*** qlstatistics: bad block in block list ***\n");
			break;
		}
		b += p->header.bytes;
#ifdef QLSTATISTICS
		if (b > qlheapmem) {
			fprintf(stderr,"too many blocks in block list\n");
			break;
		}
#endif QLSTATISTICS
	}
#endif QLCHECK
#ifdef QLSTATISTICS
#ifdef QLCHECK
	fprintf(stderr,"check: %ld blocks, %ld bytes (av. %2.2f%% bytes/block)\n",n,b,
		(n ? ((float)b / (float)n) : 0.0));
#endif QLCHECK
	{
	register struct ql_list *p = qlmisc;
	register long misc = 0, miscmem = 0, other = 0, othermem = 0;
	long	heap = qlheapmem - qlheapsize, b, qlfreemem;
	
	while (p) {
		miscmem += p->header.bytes;
		misc++;
		p = p->next;
	}
	for (b = 0 ; b < QLMAX ; b += QLALIGN)
		for (p = qllist[b / QLALIGN] ; p ; p = p->next) {
			other++;
			othermem += b;
		}
	qlfreemem = miscmem + othermem;
	
	fprintf(stderr,"blocks: %ld = %ld used + %ld free\n",
		qlused + other + misc,qlused,other + misc);
	fprintf(stderr,"bytes: %ld = %ld used + %ld free\n",
			qlfreemem + qlusedmem, qlusedmem, qlfreemem);
	fprintf(stderr,"heap: %ld bytes, %ld bytes fragmented (%2.2f%%), %ld admin bytes (%2.2f%%)\n",
			qlheapmem, heap - qlusedmem,
			(heap ? (((heap - qlusedmem) * 100.0) / heap) : 0.0),
			heap - (qlusedmem + qlfreemem),
			(heap ? (((heap - (qlusedmem + qlfreemem)) * 100.0) / heap) : 0.0));
	
	fprintf(stderr,"misc free: %ld blocks, %ld bytes (av. %2.2f bytes/block)\n",
				misc, miscmem, (misc ? ((float)miscmem / (float)misc) : 0.0));
	
	fprintf(stderr,"other free: %ld blocks, %ld bytes (av. %2.2f bytes/block)\n",
		other, othermem, (other ? ((float)othermem / (float)other) : 0.0));
	
	fprintf(stderr,"total free: %ld blocks, %ld bytes (av. %2.2f bytes/block)\n",
		other + misc, othermem + miscmem, 
		((other + misc) ? ((float)(othermem + miscmem) / (float)(other + misc)) : 0.0));
	if (qlmisc)
		fprintf(stderr,"largest free block = %ld bytes\n",(long)qlmisc->header.bytes);
#ifdef QLUSAGE
	fprintf(stderr,"%8s %8s %8s\n","size","useage","free");
	for (b = 0 ; b < QLMAX ; b += QLALIGN) {
		for (other = 0, p = qllist[b / QLALIGN] ; p ; p = p->next,other++);
		if (qluseage[b/QLALIGN])
			fprintf(stderr,"%8ld %8ld %8ld\n",b,qluseage[b/QLALIGN],other);
	}
	fprintf(stderr,"misc usage = %ld\n",qlmiscuse);
#endif QLUSAGE
	}
#endif QLSTATISTICS
}

/************************************************************************/

qlinit()
{
	if (!qllist) {
#ifdef QLCHECK
		fprintf(stderr,"initializing qlalloc routines\n");
#endif QLCHECK
		qllist = (struct ql_list **)malloc(sizeof(struct ql_list *) * (QLMAX / QLALIGN));
		memset(qllist,0,sizeof(struct ql_list *) * (QLMAX / QLALIGN));
		qlmisc = 0L;
		qlheapsize = 0;
#ifdef QLSTATISTICS
		qlheapmem = sizeof(struct ql_list *) * (QLMAX / QLALIGN);
		qlusedmem = qlused = 0;
#endif QLSTATISTICS
#ifdef QLUSAGE
		qluseage = (long *)malloc(sizeof(long) * (QLMAX / QLALIGN));
		memset(qluseage,0,sizeof(long) * (QLMAX / QLALIGN));
		qlmiscuse = 0;
#ifdef QLSTATISTICS
		qlheapmem += sizeof(long) * (QLMAX / QLALIGN);
#endif QLSTATISTICS
#endif QLUSAGE
	}
}
