/**********************************************************************
 * "@(#)interface.c 3.1 3/19/90 Sunil Sharma (Turing Institute)";
 * file: 	interface.c		(c)1989 Turing Institute
 * author: 	Sunil Sharma
 *
 * Description:
 * 	A C-Interface between SunCommonLisp and HyperNeWS.
 *	All communication to and from HyperNeWS goes through the
 *	hn_any data-structure from "../hypernews.h". The 'hnlsp_get_*'
 *	and 'hnlsp_put_*' routines build/read this data structure from
 *	lisp using the hnlsp_stack. All other interface routines use this
 *	stack to communicate their arguments/results from and to lisp.
 */

#include <stdio.h>
#include <hypernews.h>

#define	OK		"ok"
#define FAIL		"fail"
#define MAXDEPTH	64

hn_any *hnlsp_stack[MAXDEPTH];
hn_any **hnlsp_sp = hnlsp_stack;

/**********************************************************************
 * Stack operations.
 */

#define hnlsp_push(any)	(*hnlsp_sp++ = (any))
#define hnlsp_pop()	((hnlsp_sp > hnlsp_stack) ? *--hnlsp_sp : NULL)

char *hnlsp_clear()
{
	hnlsp_sp = hnlsp_stack;
	if (hn_running)
		return(OK);
	else
		return(FAIL);
}



/**********************************************************************
 * Getting data from lisp onto the stack.
 */

void hnlsp_put_null()
{
	hnlsp_push(hn_null);
}

void hnlsp_put_boolean(val)
int val;
{
	hnlsp_push(hn_new_boolean(val));
}

void hnlsp_put_integer(val)
int val;
{
	hnlsp_push(hn_new_integer(val));
}

void hnlsp_put_float(val)
float val;
{
	hnlsp_push(hn_new_number(val));
}

void hnlsp_put_string(val)
char *val;
{
	hnlsp_push(hn_new_string(val));
}

void hnlsp_put_name(val)
char *val;
{
	hnlsp_push(hn_new_name(val));
}

void hnlsp_put_message(name)
char *name;
{
	hn_any	*target = hnlsp_pop();
	hn_any *args = hnlsp_pop();

	hnlsp_push(hn_new_message(target,name,args));
}

void hnlsp_put_object(stack,type,name)
char *stack;
int type;
char *name;
{
	hnlsp_push(hn_new_object(stack,type,name));
}

void hnlsp_put_stack_object(name)
char *name;
{
	hnlsp_push(hn_new_object(NULL,hn_Stack,name));
}

void hnlsp_put_array(len)
int len;
{
	hnlsp_push(hn_new_array_body(len));
}

void hnlsp_put_array_element(n)
int n;
{
	hn_any *any = hnlsp_pop();
	hn_any *array = hnlsp_pop();

	array->u.array.argv[n] = any;
	hnlsp_push(array);
}


/**********************************************************************
 * Getting data from the stack into lisp.
 */

hnlsp_get_integer(result)
char *result;
{
        int val;
	hn_any *any = hnlsp_pop();

	if (any) {
		val = any->u.number;
		hn_free(any);
		hnlsp_set_string(result, OK);
	} else
		hnlsp_set_string(result, FAIL);
	return(val);
}

float hnlsp_get_float(result)
char *result;
{
        float val;
	hn_any *any = hnlsp_pop();

	if (any) {
		val = any->u.number;
		hn_free(any);
		hnlsp_set_string(result, OK);
	} else
		hnlsp_set_string(result, FAIL);
	return(val);
}

char *hnlsp_get_string(result)
char *result;
{
        char *val;
	hn_any *any = hnlsp_pop();

	if (any) {
		val = any->u.string;
		hn_free(any);
		hnlsp_set_string(result, OK);
	} else
		hnlsp_set_string(result, FAIL);
	return(val);
}

char *hnlsp_get_name(result)
char *result;
{
        char *val;
	hn_any *any = hnlsp_pop();

	if (any) {
		val = any->u.string;
		hn_free(any);
		hnlsp_set_string(result, OK);
	} else
		hnlsp_set_string(result, FAIL);
	return(val);
}

char *hnlsp_get_message(result)
char *result;
{
        char *val;
	hn_any *any = hnlsp_pop();

	if (any) {
		val = any->u.message.name;
		hn_free(any);
		hnlsp_set_string(result, OK);
	} else
		hnlsp_set_string(result, FAIL);
	return(val);
}

char *hnlsp_get_array_element(n)
int n;
{
	char *result;
	hn_any *any = hnlsp_pop();

	if (any) {
		if (n < any->u.array.argc) {
			hnlsp_push(any);
			hn_use(any->u.array.argv[n]);
			hnlsp_push(any->u.array.argv[n]);
			result = OK;
		}  else {
			hn_free(any);
			result = FAIL;
		}
	} else
		result = FAIL;
        return(result);
}


char *hnlsp_get_object(stack,type,name)
char *stack;
int *type;
char *name;
{
	char *result;
	hn_any *any = hnlsp_pop();

	if (any) {
		if (!any->u.object.stack)
		   hnlsp_set_string(stack, "");
                else hnlsp_set_string(stack, any->u.object.stack);
		hnlsp_set_int(type, (int)any->u.object.type);
		hnlsp_set_string(name, any->u.object.name);
		hn_free(any);
		result = OK;
	} else
		result = FAIL;
	return(result);
	}

char *hnlsp_get_type(result)
char *result;
{
	char *type;
	hn_any *any = hnlsp_pop();
	int	n;

	if (any) {
		switch ((int)any->type) {
		case hn_null_type:
			type = "null";
			break;
		case hn_number_type:
			if (any->u.number == (int)any->u.number)
				type = "integer";
			else
				type = "float";
			hnlsp_push(any);
			break;
		case hn_boolean_type:
			if (any->u.boolean)
				type = "true";
			else
				type = "false";
			break;
		case hn_string_type:
			type = "string";
			hnlsp_push(any);
			break;
		case hn_name_type:
			type = "name";
			hnlsp_push(any);
			break;
		case hn_message_type:
			type = "message";
			hn_use(any->u.message.args);
			hnlsp_push(any->u.message.args);
			hn_use(any->u.message.target);
			hnlsp_push(any->u.message.target);
			hnlsp_push(any);
			break;
		case hn_array_type:
			type = "array";
			hnlsp_push(any);
			break;
		case hn_object_type:
			type = "object";
			hnlsp_push(any);
			break;
		default:
			hnlsp_set_string(result, FAIL);
			return;
		}
		hnlsp_set_string(result, OK);
	} else {
		hnlsp_set_string(result, FAIL);
		return;}
	return (type);
	}


/**********************************************************************
 * Here are the usual interface routines.
 */

char *hnlsp_read(delay)
int delay;
{
	if (hn_ok()) {
		hnlsp_sp = hnlsp_stack;
		hnlsp_push(hn_read(delay));
		return(OK);
	} else
		return(FAIL);
}

char *hnlsp_flush()
{
	hn_flush();
	return(OK);
}

char *hnlsp_flush_input()
{
	hn_flush_input();
	return(OK);
}

char *hnlsp_ok()
{
	if (hn_ok())
		return(OK);
	else
		return(FAIL);
}

 char *hnlsp_start(name)
char *name;
{
	if (hn_start(name))
		return(OK);
	else
		return(FAIL);
}

char *hnlsp_stop()
{
	if (hn_stop())
		return(OK);
	else
		return(FAIL);
}

hnlsp_verbose(val)
int val;
{
	hn_verbose = val;
}

char *hnlsp_write()
{
	hn_any *any = hnlsp_pop();
	char *result;

	if (hn_ok()) {
		hn_write(any);
		hn_flush();
		result = OK;
	} else
		result = FAIL;
	hn_free(any);
	return(result);
}

char *hnlsp_get_any(param)
char *param;
{
	char *result;
	hn_any *target = hnlsp_pop();

	if (hn_ok()) {
		hnlsp_push(hn_get(target,param));
		result = OK;
	} else
		result = FAIL;
	hn_free(target);
	return(result);
}

char *hnlsp_put_any(param)
char *param;
{
	char *result;
	hn_any *target = hnlsp_pop();
	hn_any *value = hnlsp_pop();

	if (hn_ok()) {
		if (hn_put(target,param,value)) {
			hn_flush();
			result = OK;
		}else
			result = FAIL;
	} else
		result = FAIL;
	hn_free(target);
	hn_free(value);
	return(result);
	}


char *hnlsp_show()
{
        char *result;
	hn_any *target = hnlsp_pop();

	if (hn_ok()) {
		if (hn_show(target)) {
			hn_flush();
			result = OK;
		} else
			result = FAIL;
	} else
		result = FAIL;
	hn_free(target);
	return(result);
}

char *hnlsp_hide()
{
        char *result;
	hn_any *target = hnlsp_pop();

	if (hn_ok()) {
		if (hn_hide(target)) {
			hn_flush();
			result = OK;
		} else
			result = FAIL;
	} else
		result = FAIL;
	hn_free(target);
	return(result);
}

char *hnlsp_exists()
{
        char *result;
	hn_any *target = hnlsp_pop();

	if (hn_ok()) {
		if (hn_exists(target)) {
			result = OK;
		} else
			result = FAIL;
	} else
		result = FAIL;
	hn_free(target);
	return(result);
}

char *hnlsp_ps(postscript)
char *postscript;
{
        char *result;
	if (hn_ok()) {
		hn_ps(" ");
		if (hn_ps(postscript)) {
			hn_ps("\n");
			hn_flush();
			result = OK;
		} else
			result = FAIL;
	} else
		result = FAIL;
	return(result);
}

char *hnlsp_set_drawing(file)
char *file;
{
        char *result;
	hn_any *target = hnlsp_pop();

	if (hn_ok()) {
		if (hn_set_drawing(target,file)) {
			hn_flush();
			result = OK;
		} else
			result = FAIL;
	} else
		result = FAIL;
	hn_free(target);
	return(result);
}

char *hnlsp_rename(name)
char *name;
{
        char *result;
	hn_any *target = hnlsp_pop(), *n;

	if (hn_ok()) {
		if (n = hn_rename(target,name)) {
			hnlsp_push(n);
			result = OK;
		} else
			result = FAIL;
	} else
		result = FAIL;
	hn_free(target);
	return(result);
}

char *hnlsp_stack_copy()
{
        char *result;
	hn_any *target = hnlsp_pop(), *n;

	if (hn_ok()) {
		if (n = hn_stack_copy(target)) {
			hnlsp_push(n);
			result = OK;
		} else
			result = FAIL;
	} else
		result = FAIL;
	hn_free(target);
	return(result);
}
